/* GSTPipeline.cpp */

/* Copyright (C) 2011-2014  Lucio Carreras
 *
 * This file is part of sayonara player
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



#include "Engine/GStreamer/GSTPipeline.h"
#include <gst/app/gstappsink.h>



bool
_test_and_error(void* element, QString errorstr) {

	if(!element) {
		sp_log(LogError) << errorstr;
		return false;
	}

	return true;
}

bool
_test_and_error_bool(bool b, QString errorstr) {
	if(!b) {
		sp_log(LogError) << errorstr;
		return false;
	}

	return true;
}


void PipelineCallbacks::pad_added_handler(GstElement *src, GstPad *new_pad, gpointer data) {
	GstElement* ac = (GstElement*) data;
	GstPad* ac_pad = gst_element_get_static_pad(ac, "sink");

	if(ac_pad && gst_pad_is_linked(ac_pad)) {
        ENGINE_DEBUG << " Pads are already linked";
		return;
	}

	GstPadLinkReturn s = gst_pad_link(new_pad, ac_pad);
	if(s != GST_PAD_LINK_OK) {
        ENGINE_DEBUG << " Could not link pads";
		sp_log(LogError) << "Cannot link Error ";
		switch(s){
			case GST_PAD_LINK_WRONG_HIERARCHY:
				sp_log(LogError) << "Cause: Wrong hierarchy";
				break;

			case GST_PAD_LINK_WAS_LINKED:
				sp_log(LogError) << "Cause: Pad was already linked";
				break;
			case GST_PAD_LINK_WRONG_DIRECTION:
				sp_log(LogError) << "Cause: Pads have wrong direction";
				break;
			case GST_PAD_LINK_NOFORMAT:
				sp_log(LogError) << "Cause: Pads have incompatible format";
				break;
			case GST_PAD_LINK_NOSCHED:
				sp_log(LogError) << "Cause: Pads cannot cooperate scheduling";
				break;
			case GST_PAD_LINK_REFUSED:
			default:
				sp_log(LogError) << "Cause: Refused because of different reason";
				break;
		}
	}
}

gboolean PipelineCallbacks::show_position(gpointer data) {

	GstState state;
	GSTAbstractPipeline* pipeline = (GSTAbstractPipeline*) data;
	GstElement* p = pipeline->get_pipeline();

	if(!p) return false;
	state = pipeline->get_state();

	if( state != GST_STATE_PLAYING &&
		state != GST_STATE_PAUSED &&
		state != GST_STATE_READY)
	{
		return false;
	}

	ENGINE_DEBUG;

	pipeline->refresh_position();
	pipeline->check_about_to_finish();

	return true;
}
//#define TCP_BUFFER_SIZE 32768


#define TCP_BUFFER_SIZE 16384
static uchar data[TCP_BUFFER_SIZE];

GstFlowReturn PipelineCallbacks::new_buffer(GstElement *sink, gpointer p){

	GstSample* sample;
	GstBuffer* buffer;
	gsize size = 0;
    gsize size_new = 0;

	GSTAbstractPipeline* pipeline = (GSTAbstractPipeline*) p;

	sample = gst_app_sink_pull_sample(GST_APP_SINK(sink));
	if(!sample) {
		return GST_FLOW_OK;
	}

	buffer = gst_sample_get_buffer(sample);
	if(!buffer) {
		return GST_FLOW_OK;
	}

	size = gst_buffer_get_size(buffer);
	size_new = gst_buffer_extract(buffer, 0, data, size);
	pipeline->set_data(data, size_new);

	gst_buffer_unref(buffer);

	return GST_FLOW_OK;
}




GSTAbstractPipeline::GSTAbstractPipeline(QString name, Engine* engine, QObject* parent) :
	QObject(parent),
	SayonaraClass()
{

	_initialized = false;
	_engine = engine;
	_name = name;
    _position_ms = 0;
    _duration_ms = 0;
    _bus = 0;
    _pipeline = 0;
    _uri = NULL;
}

GSTAbstractPipeline::~GSTAbstractPipeline(){
	if (_bus){
		gst_object_unref (_bus);
	}

	if (_pipeline) {
		gst_element_set_state(GST_ELEMENT(_pipeline), GST_STATE_NULL);
		gst_object_unref (GST_OBJECT(_pipeline));
	}
}


bool GSTAbstractPipeline::init(GstState state){

	bool success = false;
	if(_initialized) {
		return true;
	}

	// create equalizer element
	_pipeline = gst_pipeline_new(_name.toStdString().c_str());
	if(!_test_and_error(_pipeline, "Engine: Pipeline sucks")){
		return false;
	}

	_bus = gst_pipeline_get_bus(GST_PIPELINE(_pipeline));

	success = create_elements();
	if(!success) {
		return false;
	}

	success = add_and_link_elements();
	if(!success) {
		return false;
	}

	configure_elements();


	gst_element_set_state(_pipeline, state);
	gst_bus_add_watch(_bus, EngineCallbacks::bus_state_changed, _engine);

	_initialized = true;
	return true;
}

void GSTAbstractPipeline::refresh_position() {

	gint64 pos;
#if GST_CHECK_VERSION(1, 0, 0)
	gst_element_query_position(GST_ELEMENT(_pipeline), GST_FORMAT_TIME, &pos);
#else
	GstFormat format = GST_FORMAT_TIME;
	gst_element_query_position(GST_ELEMENT(_pipeline), &format, &pos);
#endif

	_position_ms = pos / GST_MSECOND;
}

void GSTAbstractPipeline::refresh_duration(){

	gint64 dur;
	bool success;

#if GST_CHECK_VERSION(1, 0, 0)
	success = gst_element_query_duration(GST_ELEMENT(_pipeline), GST_FORMAT_TIME, &dur);
#else
	GstFormat format = GST_FORMAT_TIME;
	success = gst_element_query_duration(GST_ELEMENT(_pipeline), &format, &dur);
#endif

	if(success){
		_duration_ms = dur / MIO;
	}
}

void GSTAbstractPipeline::set_data(uchar* data, quint64 size){

	emit sig_data(data, size);
}

void GSTAbstractPipeline::check_about_to_finish(){

	gint64 difference = _duration_ms - _position_ms;

	if(_duration_ms >= 0){
		emit sig_pos_changed_ms((qint64) (_position_ms));
	}

	if(difference < 1000 && difference > 0 && !_about_to_finish) {

		_about_to_finish = true;
		emit sig_about_to_finish(difference);
	}

	else if(difference > 2000){
		_about_to_finish = false;
	}
}

qint64 GSTAbstractPipeline::get_duration_ms(){
    return _duration_ms;
}

qint64 GSTAbstractPipeline::get_position_ms(){
    return _position_ms;
}


void GSTAbstractPipeline::set_speed(float f) {
	Q_UNUSED(f);
}

void GSTAbstractPipeline::finished() {

	emit sig_finished();
}


GstBus* GSTAbstractPipeline::get_bus() {
	return _bus;
}


GstState GSTAbstractPipeline::get_state() {
	GstState state;
	gst_element_get_state(_pipeline, &state, NULL, 10000000);
	return state;
}


GstElement* GSTAbstractPipeline::get_pipeline() {
	return _pipeline;
}

bool GSTAbstractPipeline::set_uri(gchar* uri) {
	_uri = uri;
	return (_uri != NULL);
}

gchar* GSTAbstractPipeline::get_uri() {
	return _uri;
}


bool GSTAbstractPipeline::create_element(GstElement** elem, const gchar* elem_name, const gchar* name){

	QString error_msg;
	if(strlen(name) > 0){
		*elem = gst_element_factory_make(elem_name, name);
		error_msg = QString("Engine: ") + name + " creation failed";
	}

	else{
		*elem = gst_element_factory_make(elem_name, elem_name);
		error_msg = QString("Engine: ") + elem_name + " creation failed";
	}

	return _test_and_error(*elem, error_msg);
}
